"""
figleafsections plugin for nose.

Automatically records coverage info for Python tests and connects it with
with test was being run at the time.  Can be used to produce a "barcode"
of code execution.
"""

DEFAULT_COVERAGE_FILE='.figleaf_sections'
import pkg_resources

try:
    pkg_resources.require('figleaf>=0.6.1')
    import figleaf
except (ImportError, pkg_resources.DistributionNotFound):
    figleaf = None

import sys
err = sys.stderr

import nose.case
from nose.plugins.base import Plugin

import logging
import os

log = logging.getLogger(__name__)

def calc_testname(test):
    """
    Build a reasonably human-readable testname from each test.
    """
    name = str(test)
    if ' ' in name:
        name = name.split(' ')[1]

    return name

class FigleafSections(Plugin):
    def __init__(self):
        Plugin.__init__(self)
        self.testname = None

    def add_options(self, parser, env=os.environ):
        env_opt = 'NOSE_WITH_%s' % self.name.upper()
        env_opt.replace('-', '_')
        parser.add_option("--with-%s" % self.name,
                          action="store_true",
                          dest=self.enableOpt,
                          default=env.get(env_opt),
                          help="Enable plugin %s: %s [%s]" %
                          (self.__class__.__name__, self.help(), env_opt))

        parser.add_option("--figleaf-file",
                          action="store",
                          dest="figleaf_file",
                          default=None,
                          help="Store figleaf section coverage in this file")
        
    def configure(self, options, config):
        """
        Configure: enable plugin?  And if so, where should the coverage
        info be placed?
        """
        self.conf = config

        # enable?
        if hasattr(options, self.enableOpt):
            self.enabled = getattr(options, self.enableOpt)

        ### save coverage file name, if given.
        if options.figleaf_file:
            self.figleaf_file = options.figleaf_file
        else:
            self.figleaf_file = DEFAULT_COVERAGE_FILE

        if self.enabled and figleaf is None:
                raise Exception("You must install figleaf 0.6.1 before you can use the figleafsections plugin! See http://darcs.idyll.org/~t/projects/figleaf/doc/")

    def begin(self):
        """
        Initialize: start recording coverage info.
        """
        figleaf.start()

    def finalize(self, result):
        """
        Finalize: stop recording coverage info, save & exit.
        """
        figleaf.stop()
        
        fp = open(self.figleaf_file, 'w')
        figleaf.dump_pickled_coverage(fp)
        fp.close()

    def startTest(self, test):
        """
        Run at the beginning of each test, before per-test fixtures.

        One weakness is that this is only run for specific kinds of
        nose testcases.
        """
        if isinstance(test, nose.case.FunctionTestCase) or \
           isinstance(test, nose.case.MethodTestCase):
            self.testname = calc_testname(test)
            assert self.testname

            figleaf.start_section(self.testname)

    def stopTest(self, test):
        """
        Run at the end of each test, after per-test fixtures.
        """
        if self.testname:
            figleaf.stop_section()
            self.testname = None




"""
Stopwatch plugin for 'nose'.


"""
import sys
err = sys.stderr

from cPickle import dump, load
import time
import logging
import os
from nose.plugins.base import Plugin

log = logging.getLogger(__name__)

class Stopwatch(Plugin):
    def __init__(self):
        Plugin.__init__(self)
        self.dontrun = {}
        self.times = {}
        
    def add_options(self, parser, env=os.environ):
        Plugin.add_options(self, parser, env)
        parser.add_option("--faster-than",
                          action="store",
                          type="float",
                          dest="faster_than",
                          default=None,
                          help="Run only tests that are faster than FASTER_THAN seconds.")
        parser.add_option("--stopwatch-file",
                          action="store",
                          dest="stopwatch_file",
                          default=".nose-stopwatch-times",
                          help="Store test timing results in this file.")

    def configure(self, options, config):
        ### configure logging
        logger = logging.getLogger(__name__)
        logger.propagate = 0

        handler = logging.StreamHandler(err)
        logger.addHandler(handler)
        logger.setLevel(logging.WARNING)
        
        Plugin.configure(self, options, config)

        ### configure stopwatch stuff: file containing times, and
        ### time cutoff.

        self.stopwatch_file = os.path.abspath(options.stopwatch_file)
        self.faster_than = options.faster_than

        log.info('stopwatch module: using "%s" for stopwatch times' % \
                 (self.stopwatch_file,))
        log.info('selecting tests that run faster than %s seconds' % \
                 (self.faster_than,))

    def begin(self):
        """
        Run before any of the tests run.  Loads the stopwatch file, and
        calculates which tests should NOT be run.
        """
        try:
            self.times = load(open(self.stopwatch_file))
        except (IOError, EOFError):
            self.times = {}

        # figure out which tests should NOT be run.

        faster_than = self.faster_than
        if faster_than is not None:
            for (k, v) in self.times.items():
                if v > faster_than:
                    self.dontrun[k] = 1

    def finalize(self, result):
        """
        Save the recorded times, OR dump them into /tmp if the file
        open fails.
        """
        try:
            fp = open(self.stopwatch_file, 'w')
        except (IOError, OSError):
            t = int(time.time())
            filename = '/tmp/nose-stopwatch-%s.pickle' % (t,)
            fp = open(filename, 'w')
            log.warning('WARNING: stopwatch cannot write to "%s"' % (self.stopwatch_file))
            log.warning('WARNING: stopwatch is using "%s" to save times' % (filename,))
            
        dump(self.times, fp)
        fp.close()

    def wantMethod(self, method):
        """
        Do we want to run this method?  See _should_run.
        """
        fullname = '%s.%s.%s' % (method.__module__,
                                 method.im_class.__name__,
                                 method.__name__)

        return self._should_run(fullname)

    def wantFunction(self, func):
        """
        Do we want to run this function?  See _should_run.
        """
        fullname = '%s.%s' % (func.__module__,
                              func.__name__)

        return self._should_run(fullname)

    ### note, no wantClass, because classes only *contain* tests, but
    ### aren't tests themselves.

    def _should_run(self, name):
        """
        If we have this test listed as "don't run" because of explicit
        time constraints, don't run it.  Otherwise, indicate no preference.
        """
        if self.dontrun.has_key(name):
            return False

        return None
        
    def startTest(self, test):
        """
        startTest: start timing.
        """
        self._started = time.time()

    def stopTest(self, test):
        """
        stopTest: stop timing, canonicalize the test name, and save
        the run time.
        """
        runtime = time.time() - self._started

        # CTB: HACK!
        testname = str(test)
        if ' ' in testname:
            testname = testname.split(' ')[1]
            
        self.times[testname] = runtime



"""
decorator extension for 'nose'.

Allows you to decorate functions, classes, and methods with attributes
without modifying the actual source code.  Particularly useful in
conjunction with the 'attrib' extension package.
"""

import sys
err = sys.stderr

import logging
import os
from nose.plugins.base import Plugin

log = logging.getLogger(__name__)

def sort_plugins_by_priority(a, b):
    pa = getattr(a, 'call_priority', 100)
    pb = getattr(b, 'call_priority', 100)

    return cmp(pa, pb)

class Decorator(Plugin):
    call_priority=-100                  # put this plugin at a high priority.
    
    def __init__(self):
        Plugin.__init__(self)
        
    def add_options(self, parser, env=os.environ):
        parser.add_option("--decorator-file",
                          action="store",
                          dest="decorator_file",
                          default=None,
                          help="Apply attributes in this file to matching functions, classes, and methods")

    def configure(self, options, config):
        self.conf = config

        ### configure logging
        
        logger = logging.getLogger(__name__)
        logger.propagate = 0

        handler = logging.StreamHandler(err)
        logger.addHandler(handler)
        
        lvl = logging.WARNING
        if options.verbosity >= 5:
            lvl = 0
        elif options.verbosity >= 4:
            lvl = logging.DEBUG
        elif options.verbosity >= 3:
            lvl = logging.INFO
        logger.setLevel(lvl)

        ### enable plugin & save decorator file name, if given.
        
        if options.decorator_file:
            self.enabled = True
            self.decorator_file = options.decorator_file

    def begin(self):
        """
        Called before any tests are run.

        The only trick here is that we have to mangle the order of
        the plugins, because this plugin *must* be called before
        any plugins that examine the attributes being set.  This is
        done by sorting the plugins in-place.
        """

        ### sort plugins by specified call_priority.  HACK!
        self.conf.plugins.sort(sort_plugins_by_priority)

        ### load in the specified attributes file.
        
        filename = self.decorator_file
        
        fp = open(filename)

        curtains = {}
        for line in fp:

            # skip empty lines or lines with comments ('#')
            line = line.strip()
            if not line or line.startswith('#'):
                continue

            # parse attributes...
            name, attrib = line.split(':')
            name = name.strip()
            attrib = attrib.strip()

            # ...and store 'em.
            l = curtains.get(name, [])
            l.append(attrib)
            curtains[name] = l

        # save the attributes in 'self.curtains'.
        self.curtains = curtains

    ######
        
    def wantClass(self, cls):
        """
        wantClass -- attach matching attributes to the class.
        """
        fullname = '%s.%s' % (cls.__module__, cls.__name__,)
        self._attach_attributes(fullname, cls)

        # indicate no preferences re running this test...
        return None

    def wantMethod(self, method):
        """
        wantMethod -- attach matching attributes to this method.
        """
        fullname = '%s.%s.%s' % (method.__module__,
                                 method.im_class.__name__,
                                 method.__name__)

        self._attach_attributes(fullname, method)

        # indicate no preference re running this test...
        return None

    def wantFunction(self, func):
        """
        wantFunction -- attach matching attributes to this function.
        """
        fullname = '%s.%s' % (func.__module__,
                              func.__name__)
        
        self._attach_attributes(fullname, func)

        # indicate no preferences re running this test.
        return None

    def _attach_attributes(self, fullname, obj):
        """
        Attach attributes matching 'fullname' to the object 'obj'.
        """
        attribs = self.curtains.get(fullname, [])
        log.info('_attach_attributes: %s, %s' % (fullname, attribs,))
                  
        for a in attribs:
            obj.__dict__[a] = True
